@@ -1,3 +1,11 @@
+"""
+DetectSequence
+=====================
+Primary entry point to invoke machine learning classes in pyzm
+It is recommended you only use DetectSequence methods and not
+lower level interfaces as they may change drastically.
+"""
+
 from pyzm.helpers.Base import Base
 import pyzm.ml.object as  ObjectDetect
 import pyzm.ml.face as FaceDetect
@@ -12,7 +20,93 @@ from shapely.geometry import Polygon
 
 class DetectSequence(Base):
     def __init__(self, logger=None, options={}):
+        """Initializes ML entry point with various parameters
 
+        Args:
+            - logger (object, optional): log handler to use
+            - options (dict): Variety of ML options. Best described by an example as below
+            
+                ::
+
+                    options = {
+                        'general': {
+                            # sequence of models you want to run for every specified frame
+                            'model_sequence': 'object,face,alpr' 
+                        },
+                    
+                        # We now specify all the config parameters per model_sequence entry
+                        'object': {
+                            'general':{
+                                # 'first' - When detecting objects, if there are multiple fallbacks, break out 
+                                # the moment we get a match using any object detection library. 
+                                # 'most' - run through all libraries, select one that has most object matches
+                                # 'most_unique' - run through all libraries, select one that has most unique object matches
+                                
+                                'same_model_sequence_strategy': 'first' # 'first' 'most', 'most_unique'
+                                'pattern': '.*' # any pattern
+                            },
+
+                            # within object, this is a sequence of object detection libraries. In this case,
+                            # I want to first try on my TPU and if it fails, try GPU
+                            'sequence': [{
+                                #First run on TPU
+                                'object_weights':'/var/lib/zmeventnotification/models/coral_edgetpu/ssd_mobilenet_v2_coco_quant_postprocess_edgetpu.tflite',
+                                'object_labels': '/var/lib/zmeventnotification/models/coral_edgetpu/coco_indexed.names',
+                                'object_min_confidence': 0.3,
+                                'object_framework':'coral_edgetpu'
+                            },
+                            {
+                                # YoloV4 on GPU if TPU fails (because sequence strategy is 'first')
+                                'object_config':'/var/lib/zmeventnotification/models/yolov4/yolov4.cfg',
+                                'object_weights':'/var/lib/zmeventnotification/models/yolov4/yolov4.weights',
+                                'object_labels': '/var/lib/zmeventnotification/models/yolov4/coco.names',
+                                'object_min_confidence': 0.3,
+                                'object_framework':'opencv',
+                                'object_processor': 'gpu'
+                            }]
+                        },
+
+                        # We repeat the same exercise with 'face' as it is next in model_sequence
+                        'face': {
+                            'general':{
+                                'same_model_sequence_strategy': 'first'
+                            },
+                            'sequence': [{
+                                'face_detection_framework': 'dlib',
+                                'known_images_path': '/var/lib/zmeventnotification/known_faces',
+                                'face_model': 'cnn',
+                                'face_train_model': 'cnn',
+                                'face_recog_dist_threshold': 0.6,
+                                'face_num_jitters': 1,
+                                'face_upsample_times':1
+                            }]
+                        },
+
+                        # We repeat the same exercise with 'alpr' as it is next in model_sequence
+                        'alpr': {
+                            'general':{
+                                'same_model_sequence_strategy': 'first',
+
+                                # This can be applied to any model. This means, don't run this model
+                                # unless a previous model detected one of these labels.
+                                # In this case, I'm not calling ALPR unless we've detected a vehile
+                                # bacause platerec has an API threshold 
+
+                                'pre_existing_labels':['car', 'motorbike', 'bus', 'truck', 'boat'],
+
+                            },
+                            'sequence': [{
+                                'alpr_api_type': 'cloud',
+                                'alpr_service': 'plate_recognizer',
+                                'alpr_key': g.config['alpr_key'],
+                                'platrec_stats': 'no',
+                                'platerec_min_dscore': 0.1,
+                                'platerec_min_score': 0.2,
+                            }]
+                        }
+                    } # ml_options
+ 
+        """
         if not logger:
             logger = options.get('logger')
 
@@ -25,7 +119,7 @@ class DetectSequence(Base):
         self.media = None
         self.has_rescaled = False
 
-    def load_models(self, sequences):
+    def _load_models(self, sequences):
         if not sequences:
             sequences = self.model_sequence
         for seq in sequences:
@@ -60,7 +154,7 @@ class DetectSequence(Base):
             xfactor, yfactor, newps))
         return newps
 
-    def _filter_patterns(self,box,label,conf, global_match_pattern, polygons):
+    def _filter_patterns(self,seq, box,label,conf, polygons):
         
         # remember this needs to occur after a frame
         # is read, otherwise we don't have dimensions
@@ -74,7 +168,7 @@ class DetectSequence(Base):
                 'name': 'full_image',
                 'value': [(0, 0), (oldw, 0), (oldw, oldh), (0, oldh)],
                 #'value': [(0, 0), (5, 0), (5, 5), (0, 5)],
-                'pattern': global_match_pattern
+                'pattern': None
 
             })
             self.logger.Debug(2,'No polygons, adding full image polygon: {}'.format(polygons[0]))
@@ -91,9 +185,7 @@ class DetectSequence(Base):
                 polygons[:] = self._rescale_polygons(polygons, neww / oldw, newh / oldh)
             
 
-        global_r = re.compile(global_match_pattern)
         doesIntersect = False
-        global_match = list(filter(global_r.match, label))
         new_label = [] 
         new_bbox =[]
         new_conf = []
@@ -107,17 +199,22 @@ class DetectSequence(Base):
 
             for p in polygons:
                 poly = Polygon(p['value'])
-                self.logger.Debug(3,"intersection: comparing object:{},{} to polygon:{}".format(label[idx],obj,poly))
+                self.logger.Debug(3,"intersection: comparing object:{},{} to polygon:{},{}".format(label[idx],obj,p['name'],poly))
 
                 if obj.intersects(poly):
-                    if  p['pattern'] != global_match_pattern:
+                    if  p['pattern']:
                         self.logger.Debug(3, '{} polygon/zone has its own pattern of {}, using that'.format(p['name'],p['pattern']))
                         r = re.compile(p['pattern'])
                         match = list(filter(r.match, label))
                     else:
-                        match = global_match
-                        
-                    if label[idx].startswith('face:') or label[idx].startswith('alpr:') or label[idx] in match:
+                        match_pattern = self.ml_options.get(seq,{}).get('general',{}).get('pattern', '.*')
+                        self.logger.Debug(3,'Using global match pattern: {}'.format(match_pattern))
+
+                        r = re.compile(match_pattern)
+                        match = list(filter(r.match, label))
+
+                    #if label[idx].startswith('face:') or label[idx].startswith('alpr:') or label[idx] in match:
+                    if label[idx] in match:
                         self.logger.Debug(3,'{} intersects object:{}[{}]'.format(
                             p['name'], label[idx], b))
                         new_label.append(label[idx])
@@ -148,35 +245,44 @@ class DetectSequence(Base):
             stream (string): location of media (file, url or event ID)
             api (object): instance of the API if the stream need to route via ZM
             options (dict, optional): Various options that control the detection process. Defaults to {}:
-            - 'download': boolean # if True, will download video before analysis
-            - 'download_dir': directory where downloads will be kept (only applies to videos). Default is /tmp
-            - 'start_frame' int # Which frame to start analysis. Default 1.
-            - 'frame_skip': int # Number of frames to skip in video (example, 3 means process every 3rd frame)
-            - 'max_frames' : int # Total number of frames to process before stopping
-            - 'pattern': string # regexp for objects that will be matched. 'strategy' key below will be applied
-                         to only objects that match this pattern
-            - 'frame_set': string with exact frames to use (comma separated)
-            - 'strategy': string # various conditions to stop matching as below
-                'most_models': # Match the frame that has matched most models (does not include same model alternatives) (Default)
-                'first' # stop at first match 
-                'most' # match the frame that has the highest number of detected objects
-                'most_unique' # match the frame that has the highest number of unique detected objects
-            - 'resize': int # width to resize image, default 800
-            - 'polygons': object # set of polygons that the detected image needs to intersect
-            Returns:
-                box (array): list of bounding boxes for matched frame
-                label (array): list of labels for matched frame
-                confidence (array): list of confidences for matched frame
-                id (int): frame id of matched frame
-                img (cv2 image): image grab of matched frame
-                all_matches (array of objects): list of boxes,labels,confidences of all frames matched
 
-            Note:
-            The same frames are not retrieved depending on whether you set
-            ``download`` to ``True`` or ``False``. When set to ``True``, we use
-            OpenCV's frame reading logic and when ``False`` we use ZoneMinder's image.php function
-            which uses time based approximation. Therefore, the retrieve different frame offsets, but I assume
-            they should be reasonably close.
+                - download (boolean): if True, will download video before analysis. Defaults to False
+                - download_dir (string): directory where downloads will be kept (only applies to videos). Default is /tmp
+                - start_frame (int): Which frame to start analysis. Default 1.
+                - frame_skip: (int): Number of frames to skip in video (example, 3 means process every 3rd frame)
+                - max_frames (int): Total number of frames to process before stopping
+                - pattern (string): regexp for objects that will be matched. 'strategy' key below will be applied to only objects that match this pattern
+                - frame_set (string): comma separated frames to read. Example 'alarm,2,4,9,snapshot'
+                - strategy: (string): various conditions to stop matching as below
+               
+                    - 'most_models': Match the frame that has matched most models (does not include same model alternatives) (Default)
+                    - 'first': Stop at first match 
+                    - 'most': Match the frame that has the highest number of detected objects
+                    - 'most_unique' Match the frame that has the highest number of unique detected objects
+           
+                - resize (int): Width to resize image, default 800
+                - polygons(object): object # set of polygons that the detected image needs to intersect
+                
+        Returns:
+           - object: representing matched frame, consists of:
+
+            - box (array): list of bounding boxes for matched frame
+            - label (array): list of labels for matched frame
+            - confidence (array): list of confidences for matched frame
+            - id (int): frame id of matched frame
+            - img (cv2 image): image grab of matched frame
+
+           - array of objects:
+
+            - list of boxes,labels,confidences of all frames matched
+
+        Note:
+
+        The same frames are not retrieved depending on whether you set
+        ``download`` to ``True`` or ``False``. When set to ``True``, we use
+        OpenCV's frame reading logic and when ``False`` we use ZoneMinder's image.php function
+        which uses time based approximation. Therefore, the retrieve different frame offsets, but I assume
+        they should be reasonably close.
             
         """
 
@@ -211,7 +317,6 @@ class DetectSequence(Base):
         media = MediaStream(stream,'video', self.stream_options )
         self.media = media
 
-        global_match_pattern = self.stream_options.get('pattern', '.*')
         polygons = self.stream_options.get('polygons',[])
 
         
@@ -232,17 +337,18 @@ class DetectSequence(Base):
 
             # For each frame, loop across all models
             for seq in self.model_sequence:
+            
                 self.logger.Debug(1,'============ Frame: {} Running {} model in sequence =================='.format(self.media.get_last_read_frame(),seq))
                 pre_existing_labels = self.ml_options.get(seq,{}).get('general',{}).get('pre_existing_labels')
                 if pre_existing_labels:
                     self.logger.Debug(2,'Making sure we have matched one of {} in {} before we proceed'.format(pre_existing_labels, _labels_in_frame))
                     if not any(x in _labels_in_frame for x in pre_existing_labels):
-                        self.logger.Debug(1,'Did not find pre existing labels: {} defined, not running model'.format(pre_existing_labels))
+                        self.logger.Debug(1,'Did not find pre existing labels, not running model')
                         continue
 
                 if not self.models.get(seq):
                     try:
-                        self.load_models([seq])
+                        self._load_models([seq])
                         if manual_locking:
                             self.models[seq].acquire_lock()
                     except Exception as e:
@@ -264,31 +370,30 @@ class DetectSequence(Base):
                     cnt +=1
                     try:
                         _b,_l,_c = m.detect(image=frame)
+                        self.logger.Debug(4,'This model iteration inside {} found: labels: {},conf:{}'.format(seq, _l, _c))
+
                     except Exception as e:
                         self.logger.Error ('Error running model: {}'.format(e))
                         self.logger.Debug(2,traceback.format_exc())
-
                         continue
 
-                    self.logger.Debug(4,'This model iteration inside {} found: labels: {},conf:{}'.format(seq, _l, _c))
-                   
+                     # Now let's make sure the labels match our pattern
+                    _b,_l,_c= self._filter_patterns(seq,_b,_l,_c, polygons)
+                    if not len(_l):
+                        continue
                     if  ((same_model_sequence_strategy == 'first') 
                     or ((same_model_sequence_strategy == 'most') and (len(_l) > len(_l_best_in_same_model))) 
                     or ((same_model_sequence_strategy == 'most_unique') and (len(set(_l)) > len(set(_l_best_in_same_model))))):
                         _b_best_in_same_model = _b
                         _l_best_in_same_model = _l
                         _c_best_in_same_model = _c
-                    if (same_model_sequence_strategy=='first'):
+                    if (same_model_sequence_strategy=='first') and len(_b):
                         self.logger.Debug(3, 'breaking out of same model loop, as matches found and strategy is "first"')
                         break
                 # end of same model sequence iteration
                 # at this state x_best_in_model contains the best match across 
                 # same model variations
 
-                # Now let's make sure the labels match our pattern
-                _b_best_in_same_model,_l_best_in_same_model,_c_best_in_same_model = self._filter_patterns(_b_best_in_same_model,_l_best_in_same_model,_c_best_in_same_model, global_match_pattern, polygons)
-                if not len(_l_best_in_same_model):
-                    continue
                 _labels_in_frame.extend(_l_best_in_same_model)
                 _boxes_in_frame.extend(_b_best_in_same_model)
                 _confs_in_frame.extend(_c_best_in_same_model)
